import re
import sys

from dcplib.component_agents import IngestApiAgent
from dcplib.component_entities.ingest_entities import SubmissionEnvelope, Project

from .finder import Finder


class IngestFinder:

    name = "ingest"

    def __init__(self, deployment, **args):
        self.ingest = IngestApiAgent(deployment=deployment)

    def find(self, expression):

        field_name, field_value = expression.split('=')

        # substitute 'sub_id', 'subm_id' -> 'submission_id'
        # substitute 'proj_id', -> 'project_id'
        field_name = re.sub(r"sub([^a-z])", "submission\\1", field_name)
        field_name = re.sub(r"subm([^a-z])", "submission\\1", field_name)
        field_name = re.sub(r"proj([^a-z])", "project\\1", field_name)

        if field_name == 'submission_id':
            return SubmissionEnvelope.load_by_id(submission_id=field_value, ingest_api_agent=self.ingest)
        if field_name == 'submission_uuid':
            return self.find_submission_by_uuid(subm_uuid=field_value)
        elif field_name == 'bundle_uuid':
            return self.find_submission_with_bundle_uuid(bundle_uuid=field_value)
        elif field_name == 'project_uuid':
            return Project.load_by_uuid(project_uuid=field_value, ingest_api_agent=self.ingest)
        else:
            print(f"Sorry I don't know how to find a {field_name}")
            exit(1)

    def find_submission_by_uuid(self, subm_uuid):
        print(f"Searching for submission with UUID {subm_uuid}...")
        count = 0
        for subm in SubmissionEnvelope.iter_submissions():
            count += 1
            sys.stdout.write(f"\rSearched {count} submissions...")
            sys.stdout.flush()

            if subm.uuid == subm_uuid:
                print(f"\nSubmission {subm.envelope_id} has UUID {subm_uuid}")
                return subm

    def find_submission_with_bundle_uuid(self, bundle_uuid):
        print(f"Searching for submission with Bundle {bundle_uuid}...")
        count = 0
        for subm in SubmissionEnvelope.iter_submissions():
            count += 1
            sys.stdout.write(f"\rSearched {count} submissions...")
            sys.stdout.flush()

            for subm_bundle_uuid in subm.bundles():
                if subm_bundle_uuid == bundle_uuid:
                    print(f"\nBundle {bundle_uuid} is in the manifest for submission {subm.envelope_id}")
                    return subm


Finder.register(IngestFinder)



class Finder:

    _finders = {}

    @classmethod
    def register(cls, finder_class):
        cls._finders[finder_class.name] = finder_class

    @classmethod
    def factory(cls, finder_name, deployment, **args):
        for name, finder in cls._finders.items():
            if name == finder_name:
                return finder(deployment=deployment, **args)
        raise RuntimeError(f"Unknown finder: {finder_name}")



import requests
import json


class AzulAgent:
    def __init__(self, deployment):
        self.deployment = deployment
        if self.deployment == 'prod':
            self.azul_service_url = 'https://service.explore.data.humancellatlas.org'
        else:
            self.azul_service_url = f'https://service.{deployment}.explore.data.humancellatlas.org'

    def get_project_bundle_fqids(self, document_id, page_size=1000):
        bundle_fqids = set()

        filters = {
            'projectId': {
                'is': [
                    document_id
                ]
            }
        }
        params = {
            'filters': json.dumps(filters),
            'size': page_size
        }
        url = self.azul_service_url + f'/repository/bundles'
        page = 0

        while True:
            page += 1
            response = requests.get(url, params=params)
            response_json = response.json()
            hit_list = response_json.get('hits', [])

            for content in hit_list:
                bundle_fqids.update(f"{bundle['bundleUuid']}.{bundle['bundleVersion']}"
                                    for bundle in content['bundles'])

            pagination = response_json.get('pagination')
            if pagination is None:
                break

            search_after = pagination.get('search_after')
            search_after_uid = pagination.get('search_after_uid')

            if search_after is None and search_after_uid is None:
                break

            params.update({
                'size': page_size,
                'search_after': search_after,
                'search_after_uid': search_after_uid
            })

        return bundle_fqids



from termcolor import colored

from dcp_diag.component_entities import EntityBase


class Workflow(EntityBase):
    """ Model an analysis Workflow entity.

    A typical analysis Workflow response from the Secondary-analysis service (Cromwell API), will look like the
    following:
    {
        "end": "2019-01-06T20:35:19.533Z",
        "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
        "labels": {
            "bundle-uuid": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
            "bundle-version": "2018-11-02T114842.872218Z",
            "caas-collection-name": "xxxx-prod",
            "cromwell-workflow-id": "cromwell-aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
            "project_shortname": "project name",
            "project_uuid": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
            "workflow-name": "AdapterSmartSeq2SingleCell",
            "workflow-version": "smartseq2_v2.1.0"
        },
        "name": "AdapterSmartSeq2SingleCell",
        "start": "2019-01-06T20:18:18.102Z",
        "status": "Succeeded",
        "submission": "2019-01-06T20:16:30.804Z"
    }

    Among all of the fields, the labels is usually trivial and expensive to query for. So the default response from
    Secondary-analysis service (Cromwell API) will not contain the `labels` field unless the client is querying with
    specific parameters. This model is designed to comply with that rule, which means the `labels` is an optional
    property to this Workflow object, and will set to None is not provided. However, missing the `labels` field can
    cause  the functions that are depending on this field to run into errors.
    """

    def __init__(self, workflow_data):
        self._load_data(workflow_data)

    def __eq__(self, other):
        return (isinstance(self, type(other)) and isinstance(other, type(self)) and (
            self.uuid, self.name, self.status, self.start_time, self.end_time) == (
                other.uuid, other.name, other.status, other.start_time, other.end_time))

    def __hash__(self):
        """Calculate the hash of a subset of properties.

        Also, in Python3 world, since key lookups are always followed by the equality check, we could care less
        about the hash collisions here.
        """
        return hash((self.uuid, self.name, self.status))

    def __str__(self, prefix="", verbose=False):

        workflow_info = colored(f"{prefix}Workflow {self.uuid}\n", 'green') + \
                                f"{prefix}    name={self.name}\n" \
                                f"{prefix}    status={self.status}\n"
        if verbose:
            workflow_info += f"{prefix}    submission_time={self.submission_time}\n" \
                             f"{prefix}    start_time={self.start_time}\n" \
                             f"{prefix}    end_time={self.end_time}\n"

        if self.labels:
            workflow_info += colored(f"{prefix}    labels: \n", 'blue')
            for label_key, label_value in self.labels.items():
                if label_value:
                    workflow_info += colored(f"{prefix}        {label_key}={label_value}\n")
        return workflow_info

    def __repr__(self):
        return self.uuid

    def _load_data(self, workflow_data):
        assert isinstance(workflow_data, dict)
        self._uuid = workflow_data['id']
        self._name = workflow_data.get('name', '')
        self._status = workflow_data['status']
        self._start_time = workflow_data.get('start', '')
        self._end_time = workflow_data.get('end', '')
        self._submission_time = workflow_data.get('submission', '')

        if not workflow_data.get('labels'):
            self._labels = None
        else:
            self._labels = {
                'bundle-uuid': workflow_data['labels'].get('bundle-uuid', ''),
                'bundle-version': workflow_data['labels'].get('bundle-version', ''),
                'project_uuid': workflow_data['labels'].get('project_uuid', ''),
                'project_shortname': workflow_data['labels'].get('project_shortname', ''),
                'workflow-version': workflow_data['labels'].get('workflow-version', ''),
                'workflow-name': workflow_data['labels'].get('workflow-name', '')
            }

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def status(self):
        return self._status

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def submission_time(self):
        return self._submission_time

    @property
    def labels(self):
        return self._labels

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))
        if associated_entities_to_show:
            prefix = f"{prefix}    "

            # Show associated bundle
            # Will raise an error when the `labels` property is missing
            if any(keyword in associated_entities_to_show for keyword in ('bundles', 'bundle', 'all')):
                print(f"{prefix}Bundle:")
                print(prefix + "    " + self.labels.get('bundle-uuid'))

            # Show associated project
            # Will raise an error when the `labels` property is missing
            if any(keyword in associated_entities_to_show for keyword in ('projects', 'project', 'all')):
                print(f"{prefix}Project Shortname:")
                print(prefix + "    " + self.labels.get('project_shortname'))
                print(f"{prefix}Project UUID:")
                print(prefix + "    " + self.labels.get('project_uuid'))



from .. import DcpDiagException
from .finder import Finder

from dcp_diag.component_agents.analysis_agent import AnalysisAgent
import re


class AnalysisFinder:

    """
    dcpdig @analysis workflow_uuid=<id>
    dcpdig @analysis bundle_uuid=<id>
    """

    name = 'analysis'

    def __init__(self, deployment, **args):
        self.deployment = deployment
        # FIXME: Use a better way to authenticate instead of asking for service account JSON key
        # FIXME: If use OAuth, this should align with the Ingest Agent
        self.service_account_key = args.get('credentials')
        with AnalysisAgent.ignore_logging_msg():
            self.analysis = AnalysisAgent(deployment=self.deployment, service_account_key=self.service_account_key)

    def find(self, expression):
        """Find the target entities based on the expression.

        Args:
            expression (str): An expression representing the targeting entity with it value, separated by '='.
                It also supports shorthands, e.g.
                - "bundle_uuid='xxx'"
                - "workflow_uuid='yyy'"
                - "wf_uuid='yyy'"
        """
        if not self.service_account_key:
            raise DcpDiagException("No auth information provided, skip checking Secondary Analysis for workflows.")

        field_name, field_value = expression.split('=')

        # substitute 'wf_id', -> 'workflow_id'
        field_name = re.sub(r"wf([^a-z])", "workflow\\1", field_name)

        if field_name == 'workflow_uuid':
            print(f"Searching for workflow with UUID {field_name}...")
            with self.analysis.ignore_logging_msg():
                return self.analysis.query_by_workflow_uuid(uuid=field_value)

        elif field_name == 'bundle_uuid':
            print(f"Searching for workflow(s) with Bundle {field_name}...")
            with self.analysis.ignore_logging_msg():
                candidates = self.analysis.query_by_bundle(bundle_uuid=field_value)
                return candidates

        else:
            print(f"Sorry I don't know how to find a {field_name}")
            exit(1)


Finder.register(AnalysisFinder)



from datetime import datetime

import boto3
from botocore.errorfactory import ClientError
from sqlalchemy import create_engine, Column, Integer, String, DateTime, ForeignKey, JSON
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker, Session
from termcolor import colored

from dcplib.config import Config

from dcp_diag.component_entities import EntityBase
from .. import DcpDiagException

DbBase = declarative_base(name='DbBase')


class UploadDbConfig(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(component_name='upload', secret_name='database', **kwargs)


class DBSessionMaker:

    def __init__(self, deployment):
        engine = create_engine(UploadDbConfig(deployment=deployment).database_uri)
        DbBase.metadata.bind = engine
        self.session_maker = sessionmaker()
        self.session_maker.bind = engine

    def session(self):
        return self.session_maker()


class DbUploadArea(DbBase, EntityBase):
    __tablename__ = 'upload_area'
    id = Column(Integer(), primary_key=True)
    uuid = Column(String(), nullable=False)
    bucket_name = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    def __str__(self, prefix="", verbose=False):
        return colored(f"{prefix}Upload Area {self.id}: {self.uuid}\n", 'blue') + \
            f"{prefix}    bucket_name: {self.bucket_name}\n" \
            f"{prefix}         Status: {self.status}\n" \
            f"{prefix}     created_at: {self.created_at}\n" \
            f"{prefix}     updated_at: {self.updated_at}\n"

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))
        if associated_entities_to_show:
            if 'files' in associated_entities_to_show or 'all' in associated_entities_to_show:
                prefix = f"\t{prefix}"
                for file in self.files:
                    file.print(prefix=prefix, verbose=verbose, associated_entities_to_show=associated_entities_to_show)


class DbFile(DbBase, EntityBase):
    __tablename__ = 'file'
    id = Column(Integer(), primary_key=True)
    s3_key = Column(String(), nullable=False)
    s3_etag = Column(String(), nullable=True)
    upload_area_id = Column(Integer(), ForeignKey('upload_area.id'), nullable=False)
    name = Column(String(), nullable=False)
    size = Column(Integer(), nullable=False)
    checksums = Column(JSON(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    upload_area = relationship("DbUploadArea", back_populates="files")
    validation_files = relationship("DbValidationFiles", back_populates="file")

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}File {self.id}: {self.s3_key} with etag={self.s3_etag}\n", 'green') + \
            f"{prefix}            s3_key: {self.s3_key}\n" \
            f"{prefix}           s3_etag: {self.s3_etag}\n" \
            f"{prefix}    Upload Area ID: {self.upload_area_id}\n" \
            f"{prefix}              Name: {self.name}\n" \
            f"{prefix}              Size: {self.size}\n" \
            f"{prefix}        created_at: {self.created_at}\n" \
            f"{prefix}        updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}         checksums: {self.checksums}\n"
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):

        # Hack: keep session alive, otherwise we get:
        # Parent instance <DbFile at 0x10674a6a0> is not bound to a Session; lazy load operation of attribute
        # 'notifications' cannot proceed (Background on this error at: http://sqlalche.me/e/bhk3)
        session = Session.object_session(self)

        print(self.__str__(prefix=prefix, verbose=verbose))
        if associated_entities_to_show:
            prefix = f"\t{prefix}"
            if 'checksums' in associated_entities_to_show or 'all' in associated_entities_to_show:
                for checksum in self.checksum_records:
                    checksum.print(prefix=prefix, verbose=verbose,
                                   associated_entities_to_show=associated_entities_to_show)
            if 'validations' in associated_entities_to_show or 'all' in associated_entities_to_show:
                for join_table_row in self.validation_files:
                    val = join_table_row.validation
                    val.print(prefix=prefix, verbose=verbose,
                              associated_entities_to_show=associated_entities_to_show)
            if 'notifications' in associated_entities_to_show or 'all' in associated_entities_to_show:
                for notif in self.notifications:
                    notif.print(prefix=prefix, verbose=verbose,
                                associated_entities_to_show=associated_entities_to_show)


class DbChecksum(DbBase, EntityBase):
    __tablename__ = 'checksum'
    id = Column(String(), primary_key=True)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    job_id = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    checksum_started_at = Column(DateTime(), nullable=False)
    checksum_ended_at = Column(DateTime(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='checksum_records')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Checksum {self.id}\n", 'cyan') + \
                 f"{prefix}                File ID: {self.file_id}\n" \
                 f"{prefix}                 Job ID: {self.job_id}\n" \
                 f"{prefix}                 Status: {self.status}\n" \
                 f"{prefix}    checksum_started_at: {self.checksum_started_at}\n" \
                 f"{prefix}      checksum_ended_at: {self.checksum_ended_at}\n" \
                 f"{prefix}             created_at: {self.created_at}\n" \
                 f"{prefix}             updated_at: {self.updated_at}\n"
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))


class DbNotification(DbBase, EntityBase):
    __tablename__ = 'notification'
    id = Column(String(), primary_key=True)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    status = Column(String(), nullable=False)
    payload = Column(JSON(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='notifications')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Notification {self.id}\n", 'magenta') + \
                 f"{prefix}       File ID: {self.file_id}\n" \
                 f"{prefix}        Status: {self.status}\n" \
                 f"{prefix}    created_at: {self.created_at}\n" \
                 f"{prefix}    updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}       payload: {self.payload}\n"
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))


class DbValidation(DbBase, EntityBase):
    __tablename__ = 'validation'
    id = Column(String(), primary_key=True)
    job_id = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    results = Column(String(), nullable=False)
    validation_started_at = Column(DateTime(), nullable=False)
    validation_ended_at = Column(DateTime(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    validation_files = relationship("DbValidationFiles", back_populates="validation")

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Validation {self.id}\n", 'yellow') + \
                 f"{prefix}                   Job ID: {self.job_id}\n" \
                 f"{prefix}                   Status: {self.status}\n" \
                 f"{prefix}    validation_started_at: {self.validation_started_at}\n" \
                 f"{prefix}      validation_ended_at: {self.validation_ended_at}\n" \
                 f"{prefix}               created_at: {self.created_at}\n" \
                 f"{prefix}               updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}                  results: {self.results}\n"
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))
        if associated_entities_to_show:
            inner_prefix = f"\t{prefix}"
            if 'batch_jobs' in associated_entities_to_show or 'all' in associated_entities_to_show:
                try:
                    BatchJob.find_by_id(self.job_id).print(prefix=inner_prefix,
                                                           verbose=verbose,
                                                           associated_entities_to_show=associated_entities_to_show)
                except DcpDiagException as e:
                    print(f"{prefix}    {str(e)}\n")
                    # Batch doesn't keep records for a long time.  Proceed.


class DbValidationFiles(DbBase):
    __tablename__ = 'validation_files'
    id = Column(String(), primary_key=True)
    validation_id = Column(String(), ForeignKey('validation.id'), nullable=False)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='validation_files')
    validation = relationship("DbValidation", back_populates='validation_files')


class BatchJob(EntityBase):

    @classmethod
    def find_by_id(cls, job_id):
        batch = boto3.client('batch')
        response = batch.describe_jobs(jobs=[job_id])
        assert 'jobs' in response
        if len(response['jobs']) == 0:
            raise DcpDiagException(f"AWS does not have a record of batch job \"{job_id}\"")
        assert len(response['jobs']) == 1
        return cls(aws_job_data=response['jobs'][0])

    def __init__(self, aws_job_data):
        self.job = aws_job_data

    @property
    def id(self):
        return self.job.get('jobId')

    def __str__(self, prefix="", verbose=False):
        output = ""
        try:
            output += \
                colored(f"{prefix}Batch Job {self.id}\n", 'blue') + \
                f"{prefix}    Job Name        {self.job.get('jobName')}\n" + \
                f"{prefix}    Job Id          {self.job.get('jobId')}\n" + \
                f"{prefix}    Job Queue       {self.job.get('jobQueue')}\n" + \
                f"{prefix}    Job Definition  {self.job.get('jobDefinition')}\n" + \
                f"{prefix}    Status          {self.job.get('status')}\n" + \
                f"{prefix}    Status Reason   {self.job.get('statusReason')}\n" + \
                f"{prefix}    Created at      {self._datetime(self.job, 'createdAt')}\n" + \
                f"{prefix}    Started at      {self._datetime(self.job, 'startedAt')}\n" + \
                f"{prefix}    Stopped at      {self._datetime(self.job, 'stoppedAt')}\n"

            if 'startedAt' in self.job and 'stoppedAt' in self.job:
                duration = (self.job.get('stoppedAt') - self.job.get('startedAt')) / 1000
                output += f"{prefix}    Duration        {duration}s\n"

            output += \
                f"{prefix}    Container:\n" + \
                f"{prefix}        Image                   {self.job['container'].get('image')}\n" + \
                f"{prefix}        IvCPUs                  {self.job['container'].get('vcpus')}\n" + \
                f"{prefix}        Memory                  {self.job['container'].get('memory')}\n" + \
                f"{prefix}        Command                 {' '.join(self.job['container'].get('command'))}\n" + \
                f"{prefix}        Reason                  {self.job['container'].get('reason')}\n" + \
                f"{prefix}        Container Instance ARN  {self.job['container'].get('containerInstanceArn')}\n" + \
                f"{prefix}        Task ARN                {self.job['container'].get('taskArn')}\n" + \
                f"{prefix}        Log Stream Name         {self.job['container'].get('logStreamName')}\n"

            if 'attempts' in self.job:
                output += f"{prefix}    Attempts:\n"
                for idx, attempt in enumerate(self.job['attempts']):
                    output += \
                        f"{prefix}\tAttempt #{idx}\n" + \
                        f"{prefix}\tStarted At:     {self._datetime(attempt, 'startedAt')}\n" + \
                        f"{prefix}\tStopped At:     {self._datetime(attempt, 'stoppedAt')}\n" + \
                        f"{prefix}\tStatus Reason:  {attempt['statusReason']}\n" + \
                        f"{prefix}\tContainer:\n" + \
                        f"{prefix}\t    Container Reason:  {attempt['container'].get('reason')}\n"
        except KeyError:
            pass
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))
        if associated_entities_to_show:
            prefix = f"\t{prefix}"
            if 'logs' in associated_entities_to_show or 'all' in associated_entities_to_show:
                log_stream_name = self.job['container']['logStreamName']
                log = CloudWatchLog(log_group_name='/aws/batch/job', log_stream_name=log_stream_name)
                log.print(prefix=prefix, verbose=verbose,
                          associated_entities_to_show=associated_entities_to_show)

    @staticmethod
    def _datetime(dictionary, key):
        if key in dictionary and dictionary[key]:
            return datetime.fromtimestamp(dictionary[key]/1000).strftime('%Y-%m-%d %H:%M:%S')
        else:
            return ""


class CloudWatchLog(EntityBase):

    def __init__(self, log_group_name, log_stream_name):
        self.log_group_name = log_group_name
        self.log_stream_name = log_stream_name
        self.logs = boto3.client('logs')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Log:\n", 'red')
        if self.log_stream_name:
            try:
                response = self.logs.get_log_events(logGroupName=self.log_group_name,
                                                    logStreamName=self.log_stream_name)
                assert 'events' in response
                for event in response['events']:
                    output += event['message'] + "\n"
            except ClientError:
                pass
        else:
            output += "No log yet.\n"
        return output

    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        print(self.__str__(prefix=prefix, verbose=verbose))


DbUploadArea.files = relationship('DbFile', order_by=DbFile.id, back_populates='upload_area')
DbFile.checksum_records = relationship('DbChecksum', order_by=DbChecksum.created_at, back_populates='file')
DbFile.notifications = relationship('DbNotification', order_by=DbNotification.created_at, back_populates='file')



from cromwell_tools import api as cwm_api
from cromwell_tools import cromwell_auth as cwm_auth
from dcp_diag.component_entities.analysis_entities import Workflow
from contextlib import contextmanager
import logging


class AnalysisAgent:
    def __init__(self, deployment, service_account_key):
        """Agent model for talking to the HCA DCP Secondary-analysis service.

        This model is designed to talk to the HCA DCP Secondary-analysis service, especially the underlying Cromwell
            Workflow Execution Engine's API without exposing too many technical details to the users.

        Args:
            deployment (str): String representing the deployment environment to be queried on. It could be:
                "dev", "integration", "staging" or "prod".
            service_account_key (str): Optional, path to the service account JSON key file, which is required by
                authenticate with the Secondary-analysis service (Crowmell API). If not provided, the agent will assume
                no authentication required for talking to secondary analysis, which is very likely to break or skip
                a lot of commands that are using this agent.
                TODO: Add OAuth support to this agent so that users could authenticate through Google/Auth0.
        """
        self.deployment = deployment
        self.cromwell_url = 'https://cromwell.caas-prod.broadinstitute.org'
        self.cromwell_collection = 'lira-test' if self.deployment == 'integration' else f'lira-{self.deployment}'
        self.auth = self._get_auth(service_account_key)

    def _get_auth(self, service_account_key):
        """Helper function to generate the auth object to talk to Secondary-analysis service."""
        return cwm_auth.CromwellAuth.harmonize_credentials(service_account_key=service_account_key,
                                                           url=self.cromwell_url)

    @staticmethod
    @contextmanager
    def ignore_logging_msg(highest_level=logging.CRITICAL):
        """A context manager that will prevent any logging messages triggered during the body from being processed.

        Referred to https://gist.github.com/simon-weber/7853144
        two kind-of hacks here:
            * can't get the highest logging level in effect => delegate to the user
            * can't get the current module-level override => use an undocumented (but non-private!) interface

        highest_level (int): The maximum logging level in use. This would only need to be changed if
            a custom level greater than CRITICAL is defined.
        """
        previous_level = logging.root.manager.disable

        logging.disable(highest_level)

        try:
            yield
        finally:
            logging.disable(previous_level)

    def query_by_workflow_uuid(self, uuid):
        """Query an analysis workflow by its workflow-UUID.

        It is a one to one mapping between UUID and workflow in Secondary-analysis service (Cromwell), so the
        result should always be a single workflow. The Workflow object returned by this function will look like an
        object be loaded from the following dictionary:

            {
                "end": "2019-01-06T20:35:19.533Z",
                "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "labels": {
                    "bundle-uuid": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                    "bundle-version": "2018-11-02T114842.872218Z",
                    "caas-collection-name": "xxxx-prod",
                    "cromwell-workflow-id": "cromwell-aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                    "project_shortname": "project name",
                    "project_uuid": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                    "workflow-name": "AdapterSmartSeq2SingleCell",
                    "workflow-version": "smartseq2_v2.1.0"
                },
                "name": "AdapterSmartSeq2SingleCell",
                "start": "2019-01-06T20:18:18.102Z",
                "status": "Succeeded",
                "submission": "2019-01-06T20:16:30.804Z"
            }

        Args:
            uuid (str): Secondary-analysis service (Cromwell) workflow UUID.

        Returns:
            Workflow: Workflow object.

        Raises:
            requests.exceptions.HTTPError: When the request to Secondary-analysis service (Cromwell) failed.
        """
        query_dict = {
            'id': uuid,
            'additionalQueryResultFields': ['labels']
        }
        response = cwm_api.query(query_dict=query_dict, auth=self.auth)
        response.raise_for_status()
        result = response.json()
        all_workflows = result['results']
        total_count = result['totalResultsCount']
        assert len(all_workflows) == total_count == 1
        return Workflow(all_workflows[0])

    def query_by_bundle(self, bundle_uuid, bundle_version=None):
        """Query the analysis workflows by their workflow-UUID.

        Note, due to the open issue: https://github.com/broadinstitute/cromwell/issues/3115, if the result of workflows
        are more than ~1000, this function will very likely raise an error.

        Args:
            bundle_uuid (str): HCA DCP bundle UUID.
            bundle_version (str): Optional, HCA DCP bundle version. By default, it's None.

        Returns:
            List[Workflow]: A list of Workflow objects. E.g. [Workflow_1, ..., Workflow_100]

        Raises:
            requests.exceptions.HTTPError: When the request to Secondary-analysis service (Cromwell) failed.
        """
        query_dict = {
            'label': {
                'bundle-uuid': bundle_uuid
            },
            'additionalQueryResultFields': ['labels']
        }

        if bundle_version:
            query_dict['label']['bundle-version'] = bundle_version

        response = cwm_api.query(query_dict=query_dict, auth=self.auth)
        response.raise_for_status()
        result = response.json()
        all_workflows = [Workflow(wf) for wf in result['results']]
        return all_workflows

    def query_by_project_uuid(self, project_uuid, with_labels=True):
        """Query the analysis workflows by the HCA DCP Ingest submission project-UUID, which is essentially one of the
            workflow labels.

        Note, due to the open issue: https://github.com/broadinstitute/cromwell/issues/3115, if the result of workflows
        are more than ~1000, this function will very likely raise an error. The `with_labels` is a flag controlling the
        behavior of whether to query the workflows asking for the labels in the response, by default it's set to True,
        so please set it to False if you don't want to risk getting error responses.

        Args:
            project_uuid (str): HCA DCP Ingest submission project-UUID.
            with_labels (bool): Optional, whether to query the workflows asking for the labels in the response,
                by default it's True

        Returns:
            List[Workflow]: A list of Workflow objects. E.g. [Workflow_1, ..., Workflow_100]

        Raises:
            requests.exceptions.HTTPError: When the request to Secondary-analysis service (Cromwell) failed.
        """
        query_dict = {
            "label": {
                "project_uuid": project_uuid
            }
        }
        if with_labels:
            query_dict['additionalQueryResultFields'] = ['labels']

        response = cwm_api.query(query_dict=query_dict, auth=self.auth)
        response.raise_for_status()
        result = response.json()
        all_workflows = [Workflow(wf) for wf in result['results']]
        return all_workflows



from abc import abstractmethod


class EntityBase:

    @abstractmethod
    def __str__(self, prefix=""):
        """Return a textual representation of this entity

        :param prefix: must be output at the start of each line out output
        :return: a string

        example:

        def __str__(self, prefix=""):
          return colored(f"MyEntity {self.id}", 'red') + \
          f"{prefix}    attr1: {self.attr1}\n" + \
          f"{prefix}    attr2: {self.attr2}\n"
        """

    @abstractmethod
    def print(self, prefix="", verbose=False, associated_entities_to_show=None):
        """Display textual representation of this entity and optionally, associated ones.

        :param prefix: must be output at the start of each line out output
        :param verbose: display verbose output, pass this on to other entities
        :param associated_entities_to_show: an array of strings, each of which is an entity name, e.g. "file"
               that should also be displayed
        :return: nothing
        """



from sqlalchemy.orm.exc import NoResultFound

from .. import DcpDiagException
from .finder import Finder

from dcp_diag.component_entities.upload_entities import DbUploadArea, DbFile, DbValidation, BatchJob, DBSessionMaker


class UploadFinder:

    """
    dcpdig @upload area=<uuid>
    dcpdig @upload file_id=<upload-area>/<filename>
    """

    name = 'upload'

    def __init__(self, deployment, **args):
        self.deployment = deployment

    def find(self, expression):
        field_name, field_value = expression.split('=')
        db = DBSessionMaker(self.deployment).session()

        if field_name == 'file':
            try:
                file = db.query(DbFile).filter(DbFile.s3_key == field_value).one()
                return file
            except NoResultFound:
                raise DcpDiagException(f"No record of File \"{field_value}\""
                                       f" found in Upload's {self.deployment} database.")

        elif field_name == 'area_id' or field_name == 'area':
            try:
                area = db.query(DbUploadArea).filter(DbUploadArea.uuid == field_value).one()
                return area
            except NoResultFound:
                raise DcpDiagException(f"No record of Upload Area \"{field_value}\""
                                       f" found in Upload's {self.deployment} database.")

        elif field_name == 'validation_id' or field_name == 'validation':
            try:
                validation = db.query(DbValidation).filter(DbValidation.id == field_value).one()
                return validation
            except NoResultFound:
                raise DcpDiagException(f"No record of Validation \"{field_value}\""
                                       f" found in Upload's {self.deployment} database.")

        elif field_name == 'batch_job' or field_name == 'job_id':
            return BatchJob.find_by_id(field_value)

        else:
            raise DcpDiagException(f"Sorry I don't know how to find an {field_name}")


Finder.register(UploadFinder)
